      subroutine BondPerm
      implicit none
c
      include 'includeSystemSize.f'
      include 'includeGraphEnumeration.f'
c
      integer jBond,i,j,iGv,nGbonds,lBaseFolder,ioss,
     >     GbondsBuffer(maxBonds)
c
c      BondIndex(i,j) is a look-up table giving the bond number
c      connecting vertices i and j.
c
c      Each bond is specified by an ordered pair (i,j).  The order of
c      the pair is arbitrary, but once specified, the orientation is
c      given relative to that order.
c
c      If the canonical ordering is (i,j), then BondSign(i,j) is +1 and
c      bonsign(j,i) is -1
c
      do i=1,nVertices
         do j=1,nVertices
            BondIndex(i,j)=0
            BondSign(i,j)=0
c
            do jBond=1,nBonds
               if(Bonds(1,jBond).eq.i .and. Bonds(2,jBond).eq.j) then
                  BondIndex(i,j)=jBond
                  BondSign(i,j)=1
               else if(Bonds(2,jBond).eq.i .and.
     >                 Bonds(1,jBond).eq.j) then
                  BondIndex(i,j)=jBond
                  BondSign(i,j)=-1
               end if
            end do              ! end jBond-loop
c
         end do                 ! end j-loop
      end do                    ! end i-loop
c
      if(VERBOSE) then
         write(3,*) ' Bond look-up table'
         write(3,'("vertices",t10,"bond (sign)",t30,
     >        "vertices",t40,"bond (sign)")')
         do i = 1,nVertices
            do j=1,i
               if(BondIndex(i,j).ne.0)
     >              write(3,'(2i4,t10,i4," (",i2,")",t30,
     >              2i4,t40,i4," (",i2,")")')
     >              i,j,BondIndex(i,j),BondSign(i,j),
     >              j,i,BondIndex(j,i),BondSign(j,i)
            end do
         end do
         write(3,'(/)')
c
         write(3,*) ' These vertices should not be connected'
         write(3,'("vertices",t10,"bond (sign)",t30,
     >        "vertices",t40,"bond (sign)")')
         do i = 1,nVertices
            do j=1,i
               if(BondIndex(i,j).eq.0)
     >              write(3,'(2i4,t10,i4," (",i2,")",t30,
     >              2i4,t40,i4," (",i2,")")')
     >              i,j,BondIndex(i,j),BondSign(i,j),
     >              j,i,BondIndex(j,i),BondSign(j,i)
            end do
         end do
         write(3,'(//)')
      end if
c
c      The Gv array specifies the permutation group on vertices. A
c      permuatation group on vertices also implies an induced
c      permuation group on bonds, which is tabulated in the jGv array.
c      Bond j is carried to bond jGv(j,iGv) by symmetry operation iGv.
c
c      If the ordering of bond j, relative to the canonical ordering
c      described above, is reversed by symmetry operation iGv, then
c      iFlip(j,iGv)=-1.  If not, iFlip(j,iGv)=+1.
c
c$$$      do iGv=1,nGv
c$$$         do j=1,nBonds
c$$$c     jGv is the index of the permuted bond:
c$$$            jGv(j,iGv)=
c$$$     >           BondIndex(Gv(Bonds(1,j),iGv),Gv(Bonds(2,j),iGv))
c$$$c
c$$$c     iFlip tells whether the orientation is reversed upon permutation
c$$$c     of the bond
c$$$            iFlip(j,iGv)=
c$$$     >           BondSign(Gv(Bonds(1,j),iGv),Gv(Bonds(2,j),iGv))
c$$$            if(iFlip(j,iGv).eq.0) then
c$$$               write(*,*) ' iFlip=0 when'
c$$$               write(*,*) ' Bond ',j,' is sent to bond ',jGv(j,iGv),
c$$$     >              ' by group op ',iGv
c$$$               write(*,*) ' Bond ',j,' contains vertices ',Bonds(j,1),
c$$$     >              ' and ',Bonds(j,2)
c$$$               write(*,*) ' The new vertices are ',Gv(Bonds(1,j),iGv),
c$$$     >              ' and ',Gv(Bonds(2,j),iGv)
c$$$               stop
c$$$            end if
c$$$         end do
c$$$      end do
c
      lBaseFolder=len_trim(BaseFolder)
c
      FileName=BaseFolder(1:lBaseFolder)//'Gbonds'
      open(unit=1,file=FileName)
      nGbonds=0
      do j=1,maxGroupOrder+1
         read(1,*,iostat=ioss,end=11) (GbondsBuffer(i),i=1,nBonds)
         if(j>maxGroupOrder) go to 21
         nGbonds=nGbonds+1
         do i=1,nBonds
            jGv(i,j)=iabs(GbondsBuffer(i))
            if(GbondsBuffer(i)>0) then
               iFlip(i,j)=1
            else
               iFlip(i,j)=-1
            end if
         end do
      end do
 21   write(*,*) 'Size of Gbonds array is too small.'
      stop
 11   close(1)
c
      if(nGbonds.ne.nGv) then
         write(*,*) ' Number of input bond permutations, ',nGbonds
         write(*,*) ' differs from order of vertex group, ',nGv
         stop
      end if
c
      if(VERBOSE) then
         write(3,*) ' Permutation group on bonds'
         do iGv=1,nGv
            write(3,'(32i3)') (jGv(j,iGv),j=1,nBonds)
            write(3,'(32i3)') (iFlip(j,iGv),j=1,nBonds)
         end do
         write(3,'(//)')
      end if
c
      return
      end
