      subroutine InvariantSample(jBond,nc1)
      implicit none
c
      include 'includeSystemSize.f'
      include 'includeGraphEnumeration.f'
c
      integer :: jBond,nc1,  i,j,k,b1,b2, jA, nSample,
     >     iWork(0:maxInvariants),info,mI1,nI1,maxPop,kMinScore
c
      real*8 Dot(0:maxInvariants,0:maxInvariants),
     >     aa(0:maxInvariants,0:maxInvariants),bb(0:maxInvariants),
     >     b(0:maxInvariants),score(maxInvariants),
     >     BestScore(maxInvariants),
     >     aaf(0:maxInvariants,0:maxInvariants),scale(0:maxInvariants),
     >     cc(0:maxInvariants),work(3,0:maxInvariants),rcond,ferr,berr,
     >     proj,tol,minScore
c
      character*1 equed
c
c     InvValueList(j,i) = jth value that the ith invariant takes
c      InvValuePop(j,i) = number of configs for which the ith
c                         invariant takes the jth value
c          nInvValue(i) = number of different values taken by the ith
c                         invariant
c     "Active" invariants are the ones we will use for sorting (dealing).
c
      data tol /1.d-04/
      nSample=min(nc1,maxInvariantSampleSize)
      mI1=maxInvariants+1
c
c -- Initialize matrix of dot products
      Dot=0.d0
c
c -- Initialize best scores
      BestActiveInvList=0
      BestScore=0.d0
c
c -- Loop over sampling of configurations
      do j=1,nSample
         x=0                    ! initialize empty bonds w/ zero
         read(2) (x(k),k=1,jBond) ! read configuration
c
         do b1=1,nBonds         ! Caclulate bond products b[i]b[j] w/ i=j
            BondProduct(b1,b1)=x(b1)*x(b1)
         end do
c
         do b1=2,nBonds         ! Caclulate bond products b[i]b[j] w/ i!=j
            do b2=1,b1-1
               BondProduct(b1,b2)=x(b1)*x(b2)
               BondProduct(b2,b1)=BondProduct(b1,b2)
            end do
         end do
c
c        Calculate invariants
         InvValues(0)=1         ! the 0th invariant is always a constant
         do i=1,nInv
            InvValues(i)=0
            do k=1,lInv(i)      ! loop over terms in invariants
               InvValues(i)=InvValues(i)+BondProduct(
     >              Invariants(1,k,i),Invariants(2,k,i))*
     >              Invariants(3,k,i)
            end do              ! end loop over terms in invariant
c        Accumulate statistics on invariants
            if(j.eq.1) then     ! First config
               InvValuePop(1,i)=1
               InvValueList(1,i)=InvValues(i)
               nInvValue(i)=1
            else                ! after first config
c              Identify position of invariant value on list
               do k=1,nInvValue(i)
                  if(InvValues(i).eq.InvValueList(k,i)) then
c                    invariant value detected in previous config
                     InvValuePop(k,i)=InvValuePop(k,i)+1
                     go to 20
                  end if
               end do           ! end loop over previous values of
                                ! invariant i
c                    new value for ith invariant
               if(nInvValue(i).eq.maxInvValues) then
                  write(3,*) 'Number of values taken by invariant ',i,
     >                 ' exceeds maxInvValues.  Program halted.'
                  stop
               end if
               nInvValue(i)=nInvValue(i)+1
               InvValueList(nInvValue(i),i)=InvValues(i)
               InvValuePop(nInvValue(i),i)=1
 20            continue
            end if              ! end if-block for first config
         end do                 ! end loop over invariants
c
c -- The product of two elements in the array InvValues may be outside 
c -- the permissible range of an INTEGER*2 number.
         do i=0,nInv
            do k=0,i
               Dot(i,k)=Dot(i,k)+InvValues(i)*InvValues(k)
            end do
         end do
c
      end do                    ! end loop over config sample
      close(2)
c
      do i=1,nInv
         do k=0,i-1
            Dot(k,i)=Dot(i,k) ! Fill in other half of dot product matrix
         end do
      end do
      score=0.d0
c
c     Eventually I should figure out which half I need.
c
      write(3,'(//," -- Invariant characteristics --")')
      write(3,'("Sample of ",i8," out of total ",i8,
     >  " configs.    tol = ",e9.2)') nSample,nc1,tol
      write(3,'("inv.",t6,"norm",t15,"actv?",t21,"#vals",t27,
     >"values/population")')
c
c     Obtain a set of linearly independent invariants
      nActiveInv=0
c
      do i=1,nInv
         if(nInvValue(i)>1) then
c     Verify that the ith invariant is independent of previous
c     invariants before including it.
            bb(0)=Dot(0,i)
            aa(0,0)=Dot(0,0)
            if(nActiveInv>0) then
               do j=1,nActiveInv
                  bb(j)=Dot(ActiveInvList(j),i)
                  aa(0,j)=Dot(0,ActiveInvList(j))
                  aa(j,0)=aa(0,j)
               end do
               do j=1,nActiveInv
                  jA=ActiveInvList(j)
                  do k=1,j
                     aa(j,k)=Dot(jA,ActiveInvList(k))
                     aa(k,j)=aa(j,k)
                  end do
               end do
            end if
            do j=0,nActiveInv
               b(j)=bb(j)
            end do
            nI1=nActiveInv+1
c
c$$$            call dposvx( 'E', 'U', nI1, 1, aa, mI1,
c$$$     >           aaf, mI1, equed, scale, bb, mI1, cc, mI1,
c$$$     >           rcond, ferr, berr, work, iwork, info )
            call LapackRoutines(aa(0:nI1-1,0:nI1-1), bb(0:nI1-1),
     $           cc(0:nI1-1), nI1, nI1, info)
c
            if(info.ne.0) then
               write(3,*) 'info = ',info,' upon exit from dposvx.'
     >              ,' Program halted.'
               write(3,*) ' rcond = ',rcond
               stop
            end if
c
            proj=0.d0
            do j=0,nActiveInv
               proj=proj+b(j)*cc(j)
            end do
c$$$            proj=Dot(i,i)-proj
            proj=1.d0-proj/Dot(i,i)
c     
            if(dabs(proj)>tol) then
               nActiveInv=nActiveInv+1
               ActiveInvList(nActiveInv)=i
               maxPop=0
               do k=1,nInvValue(i)
                  if(InvValuePop(k,i).gt.maxPop) maxPop=InvValuePop(k,i)
               end do
               score(i)=nInvValue(i)+1.d0-(dfloat(maxPop)
     >              /dfloat(nSample))
c
c     Identify the weakest of the invariants currently among the
c     nInvForDeal invariants to be used for "dealing".  The weakest
c     invariant is kMinScore on the list and has score of minScore.
               minScore=1.d50
               do k=1,nInvForDeal
                  if(BestActiveInvList(k).eq.0) then
                     BestActiveInvList(k)=i
                     BestScore(k)=score(i)
                     go to 30
                  else
                     if(score(BestActiveInvList(k)).lt.minScore) then
                        minScore=score(BestActiveInvList(k))
                        kMinScore=k
                     end if
                  end if
               end do
c     If score(i) is better than minScore, replace invariant at posiiton
c     kMinScore in the list for dealing.
               if(minScore.lt.score(i)) then
                  BestActiveInvList(kMinScore)=i
                  BestScore(kMinScore)=score(i)
               end if
 30            continue
               if(VERBOSE) then
                  write(3,'(i4,e10.2,t17,"yes",t21,i3,t25,10(3x,i5))') i
     >                 ,proj,nInvValue(i),(InvValueList(k,i),k=1
     >                 ,min(10,nInvValue(i)))
                  write(3,'(1h[,f8.4,1h],t25,10i8)') score(i),
     >                 (InvValuePop(k,i),k=1,min(10,nInvValue(i)))
               end if
            else
               if(VERBOSE) then
                  write(3,'(i4,e10.2,t17," no",t21,i3,t25,10(3x,i5))') i
     >                 ,proj,nInvValue(i),(InvValueList(k,i),k=1
     >                 ,min(10,nInvValue(i)))
                  write(3,'(t25,10i8)') (InvValuePop(k,i),
     >                 k=1,min(10,nInvValue(i)))
               end if
            end if
         else
            if(VERBOSE) then
               write(3,'(i4,t18,"no",t21,i3,t29,i4)') i, nInvValue(i),
     >              (InvValueList(k,i),k=1,nInvValue(i))
               write(3,'(t25,i8)') (InvValuePop(k,i),k=1,nInvValue(i))
            end if
         end if
      end do                    ! end search for linear ind.  invariants 
c
      write(3,'(//)')
c
      write(3,*) ' Best active invariants '
      do i=1,nInvForDeal
         if(BestActiveInvList(i).eq.0) then
            nBestActiveInv=i-1
            go to 40
         else
            j=BestActiveInvList(i)
            write(3,'(i4,t21,i3,t25,10(3x,i5))') j,nInvValue(j)
     >           ,(InvValueList(k,j),k=1,min(10,nInvValue(j)))
            write(3,'(1h[,f8.4,1h],t25,10i8)') score(j),
     >           (InvValuePop(k,j),k=1,min(10,nInvValue(j)))
         end if
      end do
      nBestActiveInv=nInvForDeal
c
 40   write(3,'(//)')
c
      return
      end
c
