!#########################################################################
!  Graph Theory Routines for MkInvar: 
!     rToPrim          -- Return vertice to unit cell.
!     rToPrim_Integer  -- "   "   " and accepts integer input.
!     Sym_Multiply     -- Multiply two group operations: integer.
!     Sym_OP           -- Apply symmetry operation to vertice.
!     Sym_OPP          -- "     "     " and wrap back into cell.
!     Sym_PRIM         -- Return bond to cell
!     Sym_OPBOND       -- Apply symmetry operation to bond.
!     Sym_OPBONDP      -- "     "     " and return bond to cell.
!     Sym_IBOND        -- Identify index and sign of bond.
!     FindVertex       -- Return index of vertex within list.
!     FindBond         -- Return index of bond within list.
!     MinImage         -- Return minimum image vector between two sites.
!     BondPair_MinDist -- Return minimum distance between bond pair.
!
!#########################################################################
      subroutine rToPrim(r,Cell,CellI,r2)
      implicit none
c
      real*8 Cell(3,3),CellI(3,3),r(3),r1(3),r2(3),Mod_Positive
c
      call Matrix_Multiply_vM(r,CellI,r1)
c
      r1(1)=Mod_Positive(r1(1))
      r1(2)=Mod_Positive(r1(2))
      r1(3)=Mod_Positive(r1(3))
c
      call Matrix_Multiply_vM(r1,Cell,r2)
c
      return
      end
c
c
      subroutine rToPrim_Integer(t,Cell,CellI,t2)
      implicit none
c
      include 'param.inc'
c
      integer t(3),t2(3)
      real*8 Cell(3,3),CellI(3,3),r(3),r1(3),r2(3),Mod_Positive
c
      r(1)=dfloat(t(1))
      r(2)=dfloat(t(2))
      r(3)=dfloat(t(3))
c
      call Matrix_Multiply_vM(r,CellI,r1)
c
      r1(1)=Mod_Positive(r1(1))
      r1(2)=Mod_Positive(r1(2))
      r1(3)=Mod_Positive(r1(3))
c
      call Matrix_Multiply_vM(r1,Cell,r2)
c
      t2(1)=int(r2(1)*TOL_INTEGER)
      t2(2)=int(r2(2)*TOL_INTEGER)
      t2(3)=int(r2(3)*TOL_INTEGER)
c
      return
      end
c
c
      subroutine Sym_Multiply(G1,G2,G3,Cell,CellI)
      implicit none
c
      include 'param.inc'
c
      integer G1(3,4),G2(3,4),G3(3,4),r(3)
      real*8 Cell(3,3),CellI(3,3)
c
c -- Translation vector of new symmetry element
      call Matrix_Multiply_Mv_Integer(G1(1,1),G2(1,4),r)
c
      r(1)=r(1)+G1(1,4)
      r(2)=r(2)+G1(2,4)
      r(3)=r(3)+G1(3,4)
c
      call rToPrim_Integer(r,Cell,CellI,G3(1,4))
c
c -- Rotation matrix of new symmetry element
      call Matrix_Multiply_MN_Integer(G1(1,1),G2(1,1),G3(1,1))
c
      return
      end
c
c
      subroutine Sym_OP(G,r,r1)
      implicit none
c
      integer G(3,4),r(3),r1(3)
c
      call Matrix_Multiply_Mv_Integer(G(1,1),r,r1)
c
      r1(1)=r1(1)+G(1,4)
      r1(2)=r1(2)+G(2,4)
      r1(3)=r1(3)+G(3,4)
c
      return
      end
c
c
      subroutine Sym_OPP(G,r,r2,Cell,CellI)
      implicit none
c
      include 'param.inc'
c
      integer G(3,4),r(3),r1(3),r2(3)
      real*8 Cell(3,3),CellI(3,3)
c
      call Matrix_Multiply_Mv_Integer(G(1,1),r,r1)
c
      r1(1)=r1(1)+G(1,4)
      r1(2)=r1(2)+G(2,4)
      r1(3)=r1(3)+G(3,4)
c
      call rToPrim_Integer(r1,Cell,CellI,r2)
c
      return
      end
c
c
      subroutine Sym_PRIM(Bond1,Cell,CellI,Bond2)
      implicit none
c
      include 'param.inc'
c
      integer Bond1(3,2),Bond2(3,2),trans(3)
      real*8 Cell(3,3),CellI(3,3)
c
      call rToPrim_Integer(Bond1(1,1),Cell,CellI,Bond2(1,1))
c
      trans(1)=Bond2(1,1)-Bond1(1,1)
      trans(2)=Bond2(2,1)-Bond1(2,1)
      trans(3)=Bond2(3,1)-Bond1(3,1)
c
      Bond2(1,2)=Bond1(1,2)+trans(1)
      Bond2(2,2)=Bond1(2,2)+trans(2)
      Bond2(3,2)=Bond1(3,2)+trans(3)
c
      return
      end
c
c
      subroutine Sym_OPBOND(G,Bond1,Bond2)
      implicit none
c
      integer G(3,4),Bond1(3,2),Bond2(3,2)
c
c -- Vertex 1
      call Sym_OP(G,Bond1(1,1),Bond2(1,1))
c
c -- Vertex 2
      call Sym_OP(G,Bond1(1,2),Bond2(1,2))
c
      return
      end
c
c
      subroutine Sym_OPBONDP(G,Bond1,Cell,CellI,Bond3)
      implicit none
c
      integer G(3,4),Bond1(3,2),Bond2(3,2),Bond3(3,2)
      real*8 Cell(3,3),CellI(3,3)
c
      call Sym_OPBOND(G,Bond1,Bond2)
c
      call Sym_PRIM(Bond2,Cell,CellI,Bond3)
c
      return
      end
c
c
      subroutine Sym_IBOND(Bond,Cell_Temp,CellI_Temp,nBonds,IB)
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
c
      integer Bond(3,2),Bond1(3,2),Bond2(3,2),Bond3(3,2),IB(2),nBonds
      real*8 Cell_Temp(3,3),CellI_Temp(3,3)
c
c -- Return bond to cell
      call Sym_PRIM(Bond,Cell_Temp,CellI_Temp,Bond1)
c
c -- Search list of HBonds
      call FindBond(Bond1,nBonds,IB,HBonds)
      if(IB(2).ne.0) goto 100
c
c -- Flip bond & then return to cell
      Bond3(1,1)=Bond1(1,2)
      Bond3(2,1)=Bond1(2,2)
      Bond3(3,1)=Bond1(3,2)
      Bond3(1,2)=Bond1(1,1)
      Bond3(2,2)=Bond1(2,1)
      Bond3(3,2)=Bond1(3,1)
c
      call Sym_PRIM(Bond3,Cell_Temp,CellI_Temp,Bond2)
      call FindBond(Bond2,nBonds,IB,HBonds)
      IB(1)=-IB(1)
      if(IB(2).ne.0) goto 100
c
c -- Return vertices to cell
      call rToPrim_Integer(Bond1(1,1),Cell_Temp,CellI_Temp,Bond3(1,1))
      call rToPrim_Integer(Bond1(1,2),Cell_Temp,CellI_Temp,Bond3(1,2))
c
c -- Search list of HBonds
      call FindBond(Bond3,nBonds,IB,HBonds)
c
 100  continue
c
      return
      end
c
c
      integer function FindVertex(v,nList,VL)
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
c
      integer i,v(3),nList,VL(3,nVertSimMax)
c
      FindVertex=0
      do i=1,nList
         if((v(1).eq.VL(1,i)).and.(v(2).eq.VL(2,i))
     $        .and.(v(3).eq.VL(3,i))) then
            FindVertex=i
            goto 10
         end if
      end do
c
 10   continue
c
      return
      end
c
c
      subroutine FindBond(B,nBonds,IB,HBL)
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
c
      integer i,B(3,2),nBonds,IB(2),HBL(3,2,nBondsSimMax)
c
      IB(1)=0
      IB(2)=0
      do i=1,nBonds
         if((B(1,1)-HBL(1,1,i).eq.0)
     $        .and.(B(2,1)-HBL(2,1,i).eq.0)
     $        .and.(B(3,1)-HBL(3,1,i).eq.0)
     $        .and.(B(1,2)-HBL(1,2,i).eq.0)
     $        .and.(B(2,2)-HBL(2,2,i).eq.0)
     $        .and.(B(3,2)-HBL(3,2,i).eq.0)) then
            IB(1)=1
            IB(2)=i
            goto 100
         elseif((B(1,2)-HBL(1,1,i).eq.0)
     $        .and.(B(2,2)-HBL(2,1,i).eq.0)
     $        .and.(B(3,2)-HBL(3,1,i).eq.0)
     $        .and.(B(1,1)-HBL(1,2,i).eq.0)
     $        .and.(B(2,1)-HBL(2,2,i).eq.0)
     $        .and.(B(3,1)-HBL(3,2,i).eq.0)) then
            IB(1)=-1
            IB(2)=i
            goto 100
         end if
      end do
c
 100  continue
c
      return
      end
c
c
      subroutine MinImage(r1,r2,Cell,r3,ORTHOG)
      implicit none
c
      real*8 r1(3),r2(3),r3(3),Cell(3,3),test(3),TRANS(3),d,dist,
     $     t(3),FUNC_NORM
      integer i,j,k
      logical ORTHOG
c
c -- When lattice vectors are orthogonal
      if(ORTHOG) then
         t(1)=r2(1)-r1(1)
         t(2)=r2(2)-r1(2)
         t(3)=r2(3)-r1(3)
c
         r3(1)=t(1)-nint(t(1)/Cell(1,1))*Cell(1,1)
         r3(2)=t(2)-nint(t(2)/Cell(2,2))*Cell(2,2)
         r3(3)=t(3)-nint(t(3)/Cell(3,3))*Cell(3,3)
      else
c
c -- General algorithm for non-orthogonal lattice vectors.
c -- Vector r2 is translated by TRANS to find the shortest distance t2 r1
c -- r3 is the minimum image vector
         dist=1d6
         do i=-1,1
            t(1)=dfloat(i)
            do j=-1,1
               t(2)=dfloat(j)
               do k=-1,1
                  t(2)=dfloat(k)
                  call Matrix_Multiply_vM(t,Cell,TRANS)
c     
                  test(1)=(r2(1)+TRANS(1))-r1(1)
                  test(2)=(r2(2)+TRANS(2))-r1(2)
                  test(3)=(r2(3)+TRANS(3))-r1(3)
c     
                  d=FUNC_NORM(test)
                  if(d.lt.dist) then
                     dist=d
                     r3(1)=test(1)
                     r3(2)=test(2)
                     r3(3)=test(3)
                  end if
               end do
            end do
         end do
      end if
c     
      return
      end
c
c
      subroutine BondPair_MinDist(i,j,dist)
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
      include 'graphinvar.inc'
c
      integer i,j
      real*8 d(4),dist,BOND1(3,2),BOND2(3,2),FUNC_DIST,r(3),l
c
      l=1.d0/dfloat(LCD_VS)
c
c -- Generate physical coordinates of bond
      r(1)=dfloat(HBonds(1,1,i))*l
      r(2)=dfloat(HBonds(2,1,i))*l
      r(3)=dfloat(HBonds(3,1,i))*l
      call Matrix_Multiply_vM(r,LV_Inv,BOND1(1,1))
c
      r(1)=dfloat(HBonds(1,2,i))*l
      r(2)=dfloat(HBonds(2,2,i))*l
      r(3)=dfloat(HBonds(3,2,i))*l
      call Matrix_Multiply_vM(r,LV_Inv,BOND1(1,2))
c
      r(1)=dfloat(HBonds(1,1,j))*l
      r(2)=dfloat(HBonds(2,1,j))*l
      r(3)=dfloat(HBonds(3,1,j))*l
      call Matrix_Multiply_vM(r,LV_Inv,BOND2(1,1))
c
      r(1)=dfloat(HBonds(1,2,j))*l
      r(2)=dfloat(HBonds(2,2,j))*l
      r(3)=dfloat(HBonds(3,2,j))*l
      call Matrix_Multiply_vM(r,LV_Inv,BOND2(1,2))
c
c -- Return minimum distance of four distances
      d(1)=FUNC_DIST(BOND1(1,1),BOND2(1,1))
      d(2)=FUNC_DIST(BOND1(1,1),BOND2(1,2))
      d(3)=FUNC_DIST(BOND1(1,2),BOND2(1,1))
      d(4)=FUNC_DIST(BOND1(1,2),BOND2(1,2))
c
      dist=d(1)
      if(d(2).lt.dist) dist=d(2)
      if(d(3).lt.dist) dist=d(3)
      if(d(4).lt.dist) dist=d(4)
c
      return
      end
