!#########################################################################
!  Routines to generate graph invariants for MkInvar: 
!     Invar_Gen_B           -- Generate 1st order invariant.
!     Invar_Gen_BB          -- Generate 2nd order invariant of type 
!                               b[j]b[k].
!     Invar_Gen_BC          -- Generate 2nd order invariant of type 
!                               b[j]c[k].
!     Invar_Contract        -- Remove terms with zero coefficient from
!                               invariant.
!     Invar_Sort            -- Place terms in canonical orientation for 
!                               efficient comparison of invariants.
!     Invar_Check           -- Check if an invariant is already present
!                               in list of invariants.
!     Invar_Gen_ALL_B
!
!     Invar_Gen_ALL_BB
!
!     Invar_Gen_ALL_BC
!
!     Invar_Gen_ALL_BB_from_BC
!
! * Use of the arrays BondPerm_Mem is turned off by default.  It does lead
!   to a very fast and efficient algorithm for generating invariants, but
!   the required memory usage can easily get out of hand.  
!#########################################################################    
      subroutine Invar_Gen_B(iBOND,INV)
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
      include 'graphinvar.inc'
c
      integer iSym,iBOND,INV(3,nSym),IB(2),BOND(3,2)
      logical TEST1
c
c -- Test if bond permutation is stored in memory
      TEST1=.false.
c$$$      if(BondPerm_Mem_List(iBOND).eq.1) TEST1=.true.
c
c -- Loop over all symmetry elements
      do iSym=1,nSym
c     
c -- Bond 1
c$$$         if(TEST1) then         ! Read from memory
c$$$            IB(1)=BondPerm_Mem(1,iSym,iBOND)
c$$$            IB(2)=BondPerm_Mem(2,iSym,iBOND)
c$$$         else                   ! Calculate & store in memory
            call Sym_OPBOND(Sym(1,1,iSym),HBonds(1,1,iBOND),BOND)
            call Sym_IBOND(BOND,Cell_Inv,CellInv_Inv,nBondsInv,IB)
c$$$            BondPerm_Mem(1,iSym,iBOND)=IB(1)
c$$$            BondPerm_Mem(2,iSym,iBOND)=IB(2)
c$$$         end if
c     
c -- There is no bond #2, this is a 1st order invariant.
c -- No need to order the b[i]'s within a term.
         INV(1,iSym)=IB(1)      ! Coefficient
         INV(2,iSym)=IB(2)      ! j in b[j]
         INV(3,iSym)=0          ! k in b[k]
      end do
c
c -- Did we just store this bond in memory?
c$$$      if(.not.TEST1) BondPerm_Mem_List(iBOND)=1
c
      return
      end
c
c
      subroutine Invar_Gen_BB(iBOND,jBOND,INV)
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
      include 'graphinvar.inc'
c
      integer iSym,iBOND,jBOND,INV(3,nSym),IB(2),JB(2),BOND(3,2)
      logical TEST1,TEST2
c
c -- Test if bond permutation is stored in memory
      TEST1=.false.
      TEST2=.false.
c$$$      if(BondPerm_Mem_List(iBOND).eq.1) TEST1=.true.
c$$$      if(BondPerm_Mem_List(jBOND).eq.1) TEST2=.true.
c
c -- Loop over all symmetry elements
      do iSym=1,nSym
c     
c -- Bond 1
c$$$         if(TEST1) then         ! Read from memory
c$$$            IB(1)=BondPerm_Mem(1,iSym,iBOND)
c$$$            IB(2)=BondPerm_Mem(2,iSym,iBOND)
c$$$         else                   ! Calculate & store in memory
            call Sym_OPBOND(Sym(1,1,iSym),HBonds(1,1,iBOND),BOND)
            call Sym_IBOND(BOND,Cell_Inv,CellInv_Inv,nBondsInv,IB)
c$$$            BondPerm_Mem(1,iSym,iBOND)=IB(1)
c$$$            BondPerm_Mem(2,iSym,iBOND)=IB(2)
c$$$         end if
c     
c -- Bond 2
c$$$         if(TEST2) then         ! Read from memory
c$$$            JB(1)=BondPerm_Mem(1,iSym,jBOND)
c$$$            JB(2)=BondPerm_Mem(2,iSym,jBOND)
c$$$         else                   ! Calculate & store in memory
            call Sym_OPBOND(Sym(1,1,iSym),HBonds(1,1,jBOND),BOND)
            call Sym_IBOND(BOND,Cell_Inv,CellInv_Inv,nBondsInv,JB)
c$$$            BondPerm_Mem(1,iSym,jBOND)=JB(1)
c$$$            BondPerm_Mem(2,iSym,jBOND)=JB(2)
c$$$         end if
c
c -- Order the b[i]'s within a term.
         INV(1,iSym)=IB(1)*JB(1) ! Coefficient
         if(IB(2).le.JB(2)) then
            INV(2,iSym)=IB(2)   ! j in b[j]
            INV(3,iSym)=JB(2)   ! k in b[k]
         else
            INV(2,iSym)=JB(2)   ! j in b[j]
            INV(3,iSym)=IB(2)   ! k in b[k]
         end if
      end do
c
c -- Did we just store this bond in memory?
c$$$      if(.not.TEST1) BondPerm_Mem_List(iBOND)=1
c
      return
      end
c
c
      subroutine Invar_Gen_BC(iBOND,jBOND,INV)
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
      include 'graphinvar.inc'
c
      integer iSym,iBOND,jBOND,INV(3,nSym),IB(2),JB(2),BOND(3,2)
      logical TEST1,TEST2
c
c -- Test if bond permutation is stored in memory
      TEST1=.false.
      TEST2=.false.
c$$$      if(BondPerm_Mem_List(iBOND).eq.1) TEST1=.true.
c$$$      if(BondPerm_Mem_List(jBOND).eq.1) TEST2=.true.
c
c -- Loop over all symmetry elements
      do iSym=1,nSym
c     
c -- Bond 1
c$$$         if(TEST1) then         ! Read from memory
c$$$            IB(1)=BondPerm_Mem(1,iSym,iBOND)
c$$$            IB(2)=BondPerm_Mem(2,iSym,iBOND)
c$$$         else                   ! Calculate & store in memory
            call Sym_OPBOND(Sym(1,1,iSym),HBonds(1,1,iBOND),BOND)
            call Sym_IBOND(BOND,Cell_Inv,CellInv_Inv,nBondsInv,IB)
c$$$            BondPerm_Mem(1,iSym,iBOND)=IB(1)
c$$$            BondPerm_Mem(2,iSym,iBOND)=IB(2)
c$$$         end if
c     
c -- Bond 2
c$$$         if(TEST2) then         ! Read from memory
c$$$            JB(1)=BondPerm_Mem(1,iSym,jBOND)
c$$$            JB(2)=BondPerm_Mem(2,iSym,jBOND)
c$$$         else                   ! Calculate & store in memory
            call Sym_OPBOND(Sym(1,1,iSym),HBonds(1,1,jBOND),BOND)
            call Sym_IBOND(BOND,Cell_Inv,CellInv_Inv,nBondsInv,JB)
c$$$            BondPerm_Mem(1,iSym,jBOND)=JB(1)
c$$$            BondPerm_Mem(2,iSym,jBOND)=JB(2)
c$$$         end if
c
c -- Do not order the b[i]'s within a term.
            INV(1,iSym)=IB(1)*JB(1) ! Coefficient
            INV(2,iSym)=IB(2)   ! j in b[j]
            INV(3,iSym)=JB(2)   ! k in c[k]
      end do
c
c -- Did we just store this bond in memory?
c$$$      if(.not.TEST1) BondPerm_Mem_List(iBOND)=1
c
      return
      end
c
c
      subroutine Invar_Contract(INV,N)
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
      include 'graphinvar.inc'
c
      integer i,j,N,nTerms,INV(3,nSym)
c
c -- Reduce # of terms
      nTerms=N
      do i=2,N                  ! Loop over potentially new terms
         do j=1,i-1             ! Loop over old terms
            if((INV(2,i).eq.INV(2,j)).and.(INV(3,i).eq.INV(3,j))) then
               INV(1,j)=INV(1,j)+INV(1,i) ! Add coefficient of similar terms
               INV(1,i)=0       ! Set coefficient of duplicate term to zero
               nTerms=nTerms-1  ! Lower the # of terms by one
               goto 100
            end if
         end do
 100     continue               ! Skip over old remaining terms
c
      end do
c
c -- Remove terms that have zero as coefficient
      j=0
      do i=1,N
         if(INV(1,i).ne.0) then 
            INV(1,j+1)=INV(1,i)
            INV(2,j+1)=INV(2,i)
            INV(3,j+1)=INV(3,i)
            j=j+1
         end if
      end do
      N=j
c
      return
      end
c
c
      subroutine Invar_Sort(INV,N)
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
      include 'graphinvar.inc'
c
      integer i,j,N,nTerms,s,INV(3,nSym),count(nBondsInv),
     $     order(nBondsInv),listB(nBondsInv,nBondsInv),
     $     listC(nBondsInv,nBondsInv),place(nBondsInv,nBondsInv)
c
      if(N.eq.0) goto 110       ! No terms in invariant
c
c -- Initialization
      nTerms=0
      do i=1,nBondsInv
         count(i)=0
      end do
c
c -- Group terms with same leading bond indice
      do i=1,N
         j=INV(2,i)             ! 1st Bond Index
         count(j)=count(j)+1    ! Count # of terms
         listB(count(j),j)=INV(3,i) ! 2nd Bond Index(to be sorted)
         listC(count(j),j)=INV(1,i) ! Coefficient of term
         place(count(j),j)=i    ! Original location of term
      end do
c
c -- For each group, determine the correct order.
      do i=1,nBondsInv
         if(count(i).eq.0) goto 100 ! No terms present
         if(count(i).eq.1) then ! Only a single term present
            order(1)=1
         else                   ! Multiple terms present
            call List_Sort_Integer(listB(1,i),count(i),order)
         end if
c
c -- Place terms in correct order within each group
         do j=1,count(i)
            if(nTerms.eq.0) s=sign(1,listC(order(j),i)) ! Take first term to be positive
c
            nTerms=nTerms+1
            INV(1,nTerms)=s*listC(order(j),i) ! Coefficient
            INV(2,nTerms)=i     ! 1st Bond Index
            INV(3,nTerms)=listB(j,i) ! 2nd Bond Index
         end do

 100     continue
      end do
c
 110  continue
c
      return
      end
c
c
      subroutine Invar_Check(N,IG)
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
      include 'graphinvar.inc'
c
      integer i,j,N,IG
c
c -- Check if the last invariant is already present in the list
      IG=0
      do i=1,N-1
         if(GraphInvar_Type(i).ne.GraphInvar_Type(N)) goto 100 ! Different Type
         if(GraphInvar_nTerms(i).ne.GraphInvar_nTerms(N)) goto 100 ! Different # of terms
         do j=1,GraphInvar_nTerms(i)
            if((GraphInvar(1,j,i).ne.GraphInvar(1,j,N))
     $           .or.(GraphInvar(2,j,i).ne.GraphInvar(2,j,N))
     $           .or.(GraphInvar(3,j,i).ne.GraphInvar(3,j,N))) goto 100
         end do
         IG=i                   ! Invariants are the same
         goto 110
c
 100     continue
      end do
c
 110  continue
c
      return
      end
c
c
      subroutine Invar_Gen_ALL_B()
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
      include 'symmetry.inc'
      include 'graphinvar.inc'
c
      integer n,i,iBond,jBond,nTerms,IG,coef(nSymMax),l,GCD
c
      if(DEBUG) then 
         call DEBUG_OC(0,'Invar_Gen_ALL_B',15)
         write(LOGF,*)
         write(LOGF,*) 'Generating 1st order invariants: b[j]'
      end if
c
c -- nInv should already have been initialized
      n=nInv                    ! initial # of invariants 
      do iBond=1,nBondsInv
         nTerms=nSym
c
c -- Generate invariant
         call Invar_Gen_B(iBond,GraphInvar(1,1,nInv+1))
c
c -- Remove duplicate terms
         call Invar_Contract(GraphInvar(1,1,nInv+1),nTerms)
         GraphInvar_nTerms(nInv+1)=nTerms
         GraphInvar_Type(nInv+1)=0
c
c -- Sort Terms
         call Invar_Sort(GraphInvar(1,1,nInv+1),nTerms)
c
c -- Greatest Common Denominator of all terms including normalization
         l=nTerms+1
         do i=1,nTerms
            coef(i)=abs(GraphInvar(1,i,nInv+1))
         end do
         coef(l)=nSym           ! Old Normalization 
         call List_Union_Integer(coef,l) ! Union of list of coefficients
         call Greatest_Common_Divisor(coef,l,GCD)
c
         if(GCD.ne.1) then
            do i=1,nTerms
               GraphInvar(1,i,nInv+1)=GraphInvar(1,i,nInv+1)/GCD
            end do
         end if
c
c -- Check if invariant is new
         if(nTerms.ne.0) then
            if(nInv.eq.0) then
               IG=0
            else
               call Invar_Check(nInv+1,IG)
            end if
c
            if(IG.eq.0) then    ! New Invariant
               nInv=nInv+1
               GraphInvar_Norm(nInv)=nSym/GCD
               if((GraphInvar_Norm(nInv)*GCD).ne.nSym) then
                  write(LOGF,*) 'Error: Invar_Gen_ALL_B'
                  write(LOGF,*) 'Did not recover normalization constant'
                  write(LOGF,*) 'when multiplying by GCD.'
                  write(LOGF,*) 'Original Normalization: ',nSym
                  write(LOGF,*) 'GCD: ',GCD
                  write(LOGF,*) 'New Normalization: ',
     $                 GraphInvar_Norm(nInv)
                  stop
               end if
               GraphInvar_Pair(1,nInv)=iBond
               GraphInvar_Pair(2,nInv)=0
c
c -- Keep track of generating bond pairs for each invariant
               nBondPairs(nInv)=1
               BondPair(1,nBondPairs(nInv),nInv)=iBond
               BondPair(2,nBondPairs(nInv),nInv)=0
            else                ! Duplicate Invariant
               nBondPairs(IG)=nBondPairs(IG)+1
               if(nBondPairs(IG).gt.nBondPairsMax) then
                  write(LOGF,*) 'Error: Invar_Gen_ALL_B'
                  write(LOGF,*) 'Not enough memory allocated'
                  write(LOGF,*) '# of bond pairs found: ',
     $                 nBondPairs(IG)
                  write(LOGF,*) 'Maximum allowed(nBondPairsMax): ',
     $                 nBondPairsMax
                  stop
               end if
               BondPair(1,nBondPairs(IG),IG)=iBond
               BondPair(2,nBondPairs(IG),IG)=0
            end if
         end if
c
c -- Enough memory allocated?
         if(nInv.ge.nInvMax) then
            write(LOGF,*) 'Error: Invar_Gen_ALL_B'
            write(LOGF,*) 'Not enough memory allocated'
            write(LOGF,*) '# of invariants found(nInv): ',nInv
            write(LOGF,*) 'Maximum allowed(nInvMax):    ',nInvMax
            stop
         end if
      end do
c
      if(DEBUG) then 
         write(LOGF,*) '  # of invariants: ',nInv-n
         call DEBUG_OC(1,'Invar_Gen_ALL_B',15)
      end if
c
      return
      end
c
c
      subroutine Invar_Gen_ALL_BB()
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
      include 'symmetry.inc'
      include 'graphinvar.inc'
c
      integer n,iBond,jBond,nTerms,IG,i,coef(nSymMax),l,GCD
c
      if(DEBUG) then 
         call DEBUG_OC(0,'Invar_Gen_ALL_BB',16)
         write(LOGF,*) 
         write(LOGF,*) 'Generating 2nd order invariants: b[j]b[k]'
      end if
c
c -- nInv should already have been initialized
      n=nInv                    ! initial # of invariants 
      do iBond=1,nBondsInv
         do jBond=iBond,nBondsInv
            nTerms=nSym
c
c -- Generate invariant
            call Invar_Gen_BB(iBond,jBond,GraphInvar(1,1,nInv+1))
c
c -- Remove duplicate terms
            call Invar_Contract(GraphInvar(1,1,nInv+1),nTerms)
            GraphInvar_nTerms(nInv+1)=nTerms
            GraphInvar_Type(nInv+1)=1
c
c -- Sort Terms
            call Invar_Sort(GraphInvar(1,1,nInv+1),nTerms)
c
c -- Greatest Common Denominator of all terms including normalization
         l=nTerms+1
         do i=1,nTerms
            coef(i)=abs(GraphInvar(1,i,nInv+1))
         end do
         coef(l)=nSym           ! Old Normalization 
         call List_Union_Integer(coef,l) ! Union of list of coefficients
         call Greatest_Common_Divisor(coef,l,GCD)
c
         if(GCD.ne.1) then
            do i=1,nTerms
               GraphInvar(1,i,nInv+1)=GraphInvar(1,i,nInv+1)/GCD
            end do
         end if
c
c -- Check if invariant is new
            if(nTerms.ne.0) then
               if(nInv.eq.0) then
                  IG=0
               else
                  call Invar_Check(nInv+1,IG)
               end if
c     
               if(IG.eq.0) then ! New Invariant
                  nInv=nInv+1
                  GraphInvar_Norm(nInv)=nSym/GCD
                  if((GraphInvar_Norm(nInv)*GCD).ne.nSym) then
                     write(LOGF,*) 'Error: Invar_Gen_ALL_B'
                     write(LOGF,*) 'Did not recover normalization',
     $                    ' constant'
                     write(LOGF,*) 'when multiplying by GCD.'
                     write(LOGF,*) 'Original Normalization: ',nSym
                     write(LOGF,*) 'GCD: ',GCD
                     write(LOGF,*) 'New Normalization: ',
     $                    GraphInvar_Norm(nInv)
                     stop
                  end if
                  GraphInvar_Pair(1,nInv)=iBond
                  GraphInvar_Pair(2,nInv)=jBond
c
c -- Keep track of generating bond pairs for each invariant
                  nBondPairs(nInv)=1
                  BondPair(1,1,nInv)=iBond
                  BondPair(2,1,nInv)=jBond
               else             ! Duplicate Invariant
c$$$                  nBondPairs(IG)=nBondPairs(IG)+1
c$$$                  if(nBondPairs(IG).gt.nBondPairsMax) then
c$$$                     write(LOGF,*) 'Error: Invar_Gen_ALL_BB'
c$$$                     write(LOGF,*) 'Not enough memory allocated'
c$$$                     write(LOGF,*) '# of bond pairs found: ',
c$$$     $                    nBondPairs(IG)
c$$$                     write(LOGF,*) 'Maximum allowed(nBondPairsMax): ',
c$$$     $                    nBondPairsMax
c$$$                     stop
c$$$                  end if
c$$$                  BondPair(1,nBondPairs(IG),IG)=iBond
c$$$                  BondPair(2,nBondPairs(IG),IG)=jBond
                  if(nBondPairs(IG).lt.nBondPairsMax) then
                     nBondPairs(IG)=nBondPairs(IG)+1
                     BondPair(1,nBondPairs(IG),IG)=iBond
                     BondPair(2,nBondPairs(IG),IG)=jBond
                  end if
               end if
            end if
c
c -- Enough memory allocated?
            if(nInv.ge.nInvMax) then
               write(LOGF,*) 'Error: Invar_Gen_ALL_BB'
               write(LOGF,*) 'Not enough memory allocated'
               write(LOGF,*) '# of invariants found(nInv): ',nInv
               write(LOGF,*) 'Maximum allowed(nInvMax):    ',nInvMax
               stop
            end if
         end do
      end do
      if(DEBUG) then 
         write(LOGF,*) '  # of invariants: ',nInv-n
         call DEBUG_OC(1,'Invar_Gen_ALL_BB',16)
      end if
c
      return
      end
c
c
      subroutine Invar_Gen_ALL_BC()
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
      include 'symmetry.inc'
      include 'graphinvar.inc'
c
      integer n,iBond,jBond,nTerms,IG,i,coef(nSymMax),l,GCD
c
      if(DEBUG) then 
         call DEBUG_OC(0,'Invar_Gen_ALL_BC',16)
         write(LOGF,*) 
         write(LOGF,*) 'Generating 2nd order invariants: b[j]c[k]'
      end if
c
      n=nInv                    ! initial # of invariants 
      do iBond=1,nBondsInv
         do jBond=1,nBondsInv
            nTerms=nSym
c
c -- Generate invariant
            call Invar_Gen_BC(iBond,jBond,GraphInvar(1,1,nInv+1))
c
c -- Remove duplicate terms
            call Invar_Contract(GraphInvar(1,1,nInv+1),nTerms)
            GraphInvar_nTerms(nInv+1)=nTerms
            GraphInvar_Type(nInv+1)=2
c
c -- Sort Terms
            call Invar_Sort(GraphInvar(1,1,nInv+1),nTerms)
c
c -- Greatest Common Denominator of all terms including normalization
            l=nTerms+1
            do i=1,nTerms
               coef(i)=abs(GraphInvar(1,i,nInv+1))
            end do
            coef(l)=nSym        ! Old Normalization 
            call List_Union_Integer(coef,l) ! Union of list of coefficients
            call Greatest_Common_Divisor(coef,l,GCD)
c     
            if(GCD.ne.1) then
               do i=1,nTerms
                  GraphInvar(1,i,nInv+1)=GraphInvar(1,i,nInv+1)/GCD
               end do
            end if
c
c -- Check if invariant is new
            if(nTerms.ne.0) then
               if(nInv.eq.0) then
                  IG=0
               else
                  call Invar_Check(nInv+1,IG)
               end if
c     
               if(IG.eq.0) then ! New Invariant
                  nInv=nInv+1
                  GraphInvar_Norm(nInv)=nSym/GCD
                  if((GraphInvar_Norm(nInv)*GCD).ne.nSym) then
                     write(LOGF,*) 'Error: Invar_Gen_ALL_B'
                     write(LOGF,*) 'Did not recover normalization',
     $                    ' constant'
                     write(LOGF,*) 'when multiplying by GCD.'
                     write(LOGF,*) 'Original Normalization: ',nSym
                     write(LOGF,*) 'GCD: ',GCD
                     write(LOGF,*) 'New Normalization: ',
     $                    GraphInvar_Norm(nInv)
                     stop
                  end if
                  GraphInvar_Pair(1,nInv)=iBond
                  GraphInvar_Pair(2,nInv)=jBond
c
c -- Keep track of generating bond pairs for each invariant
                  nBondPairs(nInv)=1
                  BondPair(1,nBondPairs(nInv),nInv)=iBond
                  BondPair(2,nBondPairs(nInv),nInv)=jBond
               else             ! Duplicate Invariant
c$$$                  nBondPairs(IG)=nBondPairs(IG)+1
c$$$                  if(nBondPairs(IG).gt.nBondPairsMax) then
c$$$                     write(LOGF,*) 'Error: Invar_Gen_ALL_BC'
c$$$                     write(LOGF,*) 'Not enough memory allocated'
c$$$                     write(LOGF,*) '# of bond pairs found: ',
c$$$     $                    nBondPairs(IG)
c$$$                     write(LOGF,*) 'Maximum allowed(nBondPairsMax): ',
c$$$     $                    nBondPairsMax
c$$$                     stop
c$$$                  end if
c$$$                  BondPair(1,nBondPairs(IG),IG)=iBond
c$$$                  BondPair(2,nBondPairs(IG),IG)=jBond
                  if(nBondPairs(IG).lt.nBondPairsMax) then
                     nBondPairs(IG)=nBondPairs(IG)+1
                     BondPair(1,nBondPairs(IG),IG)=iBond
                     BondPair(2,nBondPairs(IG),IG)=jBond
                  end if
               end if
            end if
c
c -- Enough memory allocated?
            if(nInv.ge.nInvMax) then
               write(LOGF,*) 'Error: Invar_Gen_ALL_BC'
               write(LOGF,*) 'Not enough memory allocated'
               write(LOGF,*) '# of invariants found(nInv): ',nInv
               write(LOGF,*) 'Maximum allowed(nInvMax):    ',nInvMax
               stop
            end if
         end do
      end do
c
      if(DEBUG) then 
         write(LOGF,*) '  # of invariants: ',nInv-n
         call DEBUG_OC(1,'Invar_Gen_ALL_BC',16)
      end if
c
      return
      end
c
c
      subroutine Invar_Gen_ALL_BB_from_BC()
      implicit none
c
c
      include 'param.inc'
      include 'structure.inc'
      include 'symmetry.inc'
      include 'graphinvar.inc'
c
      integer i,j,k,iInv,b1,b2,n,nStart,nStop,iBond,jBond,nTerms,IG,
     $     coef(nSymMax),l,GCD
      logical test
c
      if(DEBUG) then 
         call DEBUG_OC(0,'Invar_Gen_ALL_BB_from_BC',24)
         write(LOGF,*) 
         write(LOGF,*) 'Generating 2nd order invariants: b[j]b[k]'
      end if
c
      n=nInv                    ! initial # of invariants
c
c -- What range of invariant indices contain b[j]c[k] type invariants?
c -- All b[j]c[k] invariants should be contiguous in the array.
      nStart=0
      nStop=0
      do i=1,nInv
         if((GraphInvar_Type(i).eq.2).and.(nStart.eq.0)) nStart=i
         if((GraphInvar_Type(i).eq.2).and.(nStart.ne.0)) nStop=i
      end do
c
      if(DEBUG) then
         write(LOGF,*)
         write(LOGF,*) 'Range of indices for b[j]c[k] invariants: ',
     $        nStart,'  ',nStop
      end if
c
c -- Loop over b[j]c[k] invariants
      do iInv=nStart,nStop
         nTerms=GraphInvar_nTerms(iInv)
         GraphInvar_Type(nInv+1)=1
c
c -- Replace c[k] -> b[k] 
         do i=1,nTerms
            GraphInvar(1,i,nInv+1)=GraphInvar(1,i,iInv)
            b1=GraphInvar(2,i,iInv)
            b2=GraphInvar(3,i,iInv)
c
            if(b1.le.b2) then
               GraphInvar(2,i,nInv+1)=b1
               GraphInvar(3,i,nInv+1)=b2
            else
               GraphInvar(2,i,nInv+1)=b2
               GraphInvar(3,i,nInv+1)=b1
            end if
         end do
c
c -- Remove duplicate terms
         call Invar_Contract(GraphInvar(1,1,nInv+1),nTerms)
         GraphInvar_nTerms(nInv+1)=nTerms
c
c -- Sort Terms
         call Invar_Sort(GraphInvar(1,1,nInv+1),nTerms)
c
c -- Greatest Common Denominator of all terms including normalization
         l=nTerms+1
         do i=1,nTerms
            coef(i)=abs(GraphInvar(1,i,nInv+1))
         end do
         coef(l)=GraphInvar_Norm(iInv) ! Old Normalization 
         call List_Union_Integer(coef,l) ! Union of list of coefficients
         call Greatest_Common_Divisor(coef,l,GCD)
c
         if(GCD.ne.1) then
            do i=1,nTerms
               GraphInvar(1,i,nInv+1)=GraphInvar(1,i,nInv+1)/GCD
            end do
         end if
c
c -- Check if invariant is new
            if(nTerms.ne.0) then
               call Invar_Check(nInv+1,IG)
c     
               if(IG.eq.0) then ! New Invariant
                  nInv=nInv+1
                  GraphInvar_Norm(nInv)=GraphInvar_Norm(iInv)/GCD
                  if((GraphInvar_Norm(nInv)*GCD).ne.
     $                 GraphInvar_Norm(iInv)) then
                     write(LOGF,*) 'Error: Invar_Gen_ALL_BB_from_BC'
                     write(LOGF,*) 'Did not recover normalization',
     $                    ' constant'
                     write(LOGF,*) 'when multiplying by GCD.'
                     write(LOGF,*) 'Original Normalization: ',
     $                    GraphInvar_Norm(iInv)
                     write(LOGF,*) 'GCD: ',GCD
                     write(LOGF,*) 'New Normalization: ',
     $                    GraphInvar_Norm(nInv)
                     stop
                  end if
                  GraphInvar_Pair(1,nInv)=GraphInvar_Pair(1,iInv)
                  GraphInvar_Pair(2,nInv)=GraphInvar_Pair(2,iInv)
c
c -- Keep track of generating bond pairs for each invariant
                  nBondPairs(nInv)=nBondPairs(iInv)
                  j=0
                  do i=1,nBondPairs(nInv)
                     b1=BondPair(1,i,iInv)
                     b2=BondPair(2,i,iInv)
                     if(b1.le.b2) then
                        BondPair(1,j+1,nInv)=b1
                        BondPair(2,j+1,nInv)=b2
                     else
                        BondPair(1,j+1,nInv)=b2
                        BondPair(2,j+1,nInv)=b1
                     end if
c
c -- Remove duplicate bond pairs
                     if(j.eq.0) then
                        test=.true.
                     else
                        do k=1,j
                           if((BondPair(1,k,nInv).eq.
     $                          BondPair(1,j+1,nInv)).and.
     $                          (BondPair(2,k,nInv).eq.
     $                          BondPair(2,j+1,nInv))) goto 100
                        end do
                        test=.true.
                     end if
                     if(test) j=j+1
 100                 continue   ! Bond pairs are the same
                     nBondPairs(nInv)=j
                  end do
               end if
            end if
c
c -- Enough memory allocated?
            if(nInv.ge.nInvMax) then
               write(LOGF,*) 'Error: Invar_Gen_ALL_BB_from_BC'
               write(LOGF,*) 'Not enough memory allocated'
               write(LOGF,*) '# of invariants found(nInv): ',nInv
               write(LOGF,*) 'Maximum allowed(nInvMax):    ',nInvMax
               stop
            end if
      end do
c
      if(DEBUG) then 
         write(LOGF,*) '  # of invariants: ',nInv-n
         call DEBUG_OC(1,'Invar_Gen_ALL_BB_from_BC',24)
      end if
c
      return
      end
