!#########################################################################
!  MkInvar:   
!     Generate symmetry group, vertices, H-bonds, and invariants 
!      for crystal system.
!
!  To compile:
!             make
!
!  To Execute:
!             MkInvar input.txt
!
!  Compilation and Execution successful with the following compilers:
!              g77   (gcc version 3.4.6 20060404 (Red Hat 3.4.6-4))
!              pgf77, pgf90 (7.2-1 64-bit target on x86-64 Linux -tp k8-64)
!
!  I/O: - needs updating
!              1. Input File(Read_Input)
!              2. Space Group File(Read_Input)
!             
!  Input file:
!              Simulation cell dimensions
!              Cell Vectors
!              Space Group index
!              Path to space group file
!              Lattice parameters
!              Maximum allowed H-bond OO distance
!              Method to generate unit cell structure
!              Generate invariants?
!              
!  Current Work:
!     Cleaning Code
!     Conform to f90 Standard?
!     Use output from enumeration program to generate xyz configs.
!
!
!  Still to do:   
!     1. Cover Cell Invariants
!        a. Remove algebraic linear dependencies among invariants.
!     2. Test for consistency among orbits.
!     3. Replace duplicate chunks of code w/ sub-routines.
!     4. Condense overlapping arrays into single array.
!     5. Using BondPerm_Mem requires care regarding large memory usage.
!         Probably only useful for large simulation cells when only
!         several invariants need to be generated.
!     6. Confirm calculation is independent of units
!            
!
!  Version: 05_23_2009
!#########################################################################
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
c
c -- Initialization
      SIMCELL=.false.
      COVERCELL=.false.
      TERMINATE=.false.
c
      open(unit=LOGF,file='MkInvar.log')
c
c -- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c -- Read input files
c -- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      call Read_Input()
c
c -- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c -- Generate group of simulation cell
c -- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      call GenGroup()
c
c -- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c -- Generate vertices & bonds for unit, sim., and cov. cell
c -- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      call GenStructure()
c
c -- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c -- Generate invariants for cell of interest(unit, sim, or cov)
c -- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      if(jobGENINVAR.ne.0) call GenInvariant()
c
c -- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c -- Generate input for enumeration code(unit, sim, or cov)
c -- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      call Write_Enumeration_Input()
c
c -- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c -- Generate sim. cell invariants from cov. cell invariants
c -- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      if(COVERCELL) call GenInvariantCover()
c
c -- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c -- Run enumeration code
c -- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

c
c -- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c -- Generate xyz configs using output from enumeration code
c -- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

c
c -- Print estimate for memory usage
c$$$      if(DEBUG) call MemoryEstimate()
c
c -- Code exits successfully
      if(DEBUG) then
         write(LOGF,*)
         write(LOGF,*) '**************************'
         write(LOGF,*) 'MkInvar exits successfully'
         write(LOGF,*) '**************************'
      end if
c
      close(LOGF)
c
      stop
      end
