      implicit none
c
      integer max
      parameter(max=5000)
c
      integer i,j,N,nRow,nCol,LWORK
      real*8 Matrix(max,max)
      real t1,t2,elapsed(2)
c
c -- Read matrix from file
      read(*,*) nRow,nCol
      do i=1,nRow
         read(*,*) (Matrix(i,j),j=1,nCol)
      end do
c
      N=nRow
      if(N.lt.nCol) N=nCol
      LWORK=5*N*15
      t1=etime(elapsed)
c
      call LAPACK_NULLSPACE_SVD(Matrix,max,N,LWORK,nRow,nCol)
c
      t2=etime(elapsed)
      write(*,*)
      write(*,*) 'End: total = ',t2-t1,' user = ',elapsed(1),
     $     ' system = ',elapsed(2)
c
c -- Write null space to file
      open(unit=1,file='Null.txt')
      do i=1,N
         write(1,10) (Matrix(i,j),j=1,nCol)
      end do
      close(1)
c
 10   format(1x,5000(f12.8,1x))
c
      stop
      end
c
c
      subroutine LAPACK_NULLSPACE_SVD(M,max,N,LWORK,NR,NC)
      implicit none
c
      integer N,NR,NC,LWORK,max
      real*8 TOL_ZERO
      parameter(TOL_ZERO=1.d-8)
c
      integer i,j,LDA,INFO,nZero,min,nBasis
      real*8 M(max,max),WORK(LWORK),S(N),U(NR,NR),VT(NC,NC)
c
      min=NR
      if(min.gt.NC) min=NC
c
c -- Write matrix
c$$$      write(*,*)
c$$$      write(*,*) 'Original Matrix: ',NR,' x ',NC
c$$$      do i=1,NR
c$$$         write(*,10) (M(i,j),j=1,NC)
c$$$      end do
c
c -- Compute SVD: PG 146
      call DGESVD('N','A',NR,NC,M,max,S,U,NR,VT,NC,WORK,LWORK,INFO)
      write(*,*) 'Optimal Block Size: ',WORK(1)
      write(*,*) 'LWORK:              ',LWORK
c
c -- Write Singular Values
c$$$      write(*,*)
c$$$      write(*,*) 'Singular Values: ',min
c$$$      do i=1,min
c$$$         write(*,*) 'i: ',i,'  S(i): ',S(i)
c$$$      end do
c
c -- Count zeroes
      nZero=0
      do i=1,min
         if(S(i).le.TOL_ZERO) nZero=nZero+1
      end do
      nBasis=abs(NR-NC)+nZero
      write(*,*)
      write(*,*) '# of zeroes:        ',nZero
      write(*,*) '# of basis vectors: ',nBasis
c
c -- Write Singular Vectors(U)
c$$$      write(*,*)
c$$$      write(*,*) 'Singular Vectors(U) Matrix: ',NR,' x ',NR
c$$$      do i=1,NR
c$$$         write(*,10) (U(i,j),j=1,NR)
c$$$      end do
c
c -- Write Singular Vectors(VT)
c$$$      write(*,*)
c$$$      write(*,*) 'Singular Vectors(VT) Matrix: ',NC,' x ',NC
c$$$      do i=1,NC
c$$$         write(*,10) (VT(i,j),j=1,NC)
c$$$      end do
c
c -- Overwrite input(that was destroyed) with output
      N=nBasis
      do i=1,N
         do j=1,NC
            M(i,j)=VT(NC-N+i,j)
         end do
      end do
c
c -- Orthogonal Basis for Null Space
c$$$      write(*,*)
c$$$      write(*,*) 'Basis for Null Space'
c$$$      do i=1,N
c$$$         write(*,10) (M(i,j),j=1,NC)
c$$$      end do
c
 10   format(1x,10(f8.4,2x))
c
      return
      end
