!#########################################################################
!  HBondConfigXYZ:   
!     Generate cartesian coordinates for H-bond configurations given 
!      bond variables and structural information about vertices & bonds.
!
!  To compile:
!             f77 HBondConfigXYZ.f -I HBondConfigXYZ.inc -o HBondConfigXYZ
!
!  To Execute:
!             HBondConfigXYZ < Structure.txt
!
!  Compilation and Execution successful with the following compilers:
!              g77   (gcc version 3.4.6 (Ubuntu 3.4.6-6ubuntu5))
!
!  I/O:
!      1. XFile(Bond Variables)
!      *. Structure.txt(Std. Inp.)
!      2. HBondConfig.xyz
!             
!              
!  Current Work:
!
!
!  Still to do:   
!     1. 
!            
!
!  Version: 04_19_2009
!#########################################################################
      implicit none
c
      include 'HBondConfigXYZ.inc'
c
      integer i
c
c -- Read Structure.txt file
      call ReadStructure()
c
c -- Generate translations for PBC of hydrogens
      call GenerateTRANS()
c
c -- Read Bond Variables from XFile and generate cartesian coordinates
      open(unit=1,file='XFile')
      open(unit=2,file='HBondConfig.xyz')
c
      nConfig=0
      do while(.true.)
         read(1,*,end=99) (b(i),i=1,nBonds)
         nConfig=nConfig+1
         call GenerateHYD()
         call PBCHYD()
         call WriteXYZ()
      end do
 99   continue
c
      close(1)
      close(2)
c
      stop
      end
c
c
      subroutine ReadStructure()
      implicit none
c
      include 'HBondConfigXYZ.inc'
c
      integer i,j,list(12)
      real*8 r(3)
c
      read(*,*) (LatCart(1,j),j=1,3) ! We use these vectors to generate
      read(*,*) (LatCart(2,j),j=1,3) ! cartesian coordinates.
      read(*,*) (LatCart(3,j),j=1,3)
c
      read(*,*) (LatPBC(1,j),j=1,3) ! We use these vectors for periodic
      read(*,*) (LatPBC(2,j),j=1,3) ! boundary conditions.
      read(*,*) (LatPBC(3,j),j=1,3)
c
      read(*,*) nVertices
      if(nVertices.gt.nBondsMax/2) then
         write(*,*) 'Error: ReadStructure'
         write(*,*) 'Not enough memory allocated.'
         write(*,*) 'nVertMax:  ',nBondsMax/2
         write(*,*) 'nVertices: ',nVertices
      end if
c
      do i=1,nVertices
         read(*,*) (list(j),j=1,6)
         r(1)=dfloat(list(1))/dfloat(list(2))
         r(2)=dfloat(list(3))/dfloat(list(4))
         r(3)=dfloat(list(5))/dfloat(list(6))
         call Matrix_Multiply_vM(r,LatCart,Vertice(1,i))
      end do
c
      read(*,*) nBonds
      if(nBonds.gt.nBondsMax) then
         write(*,*) 'Error: ReadStructure'
         write(*,*) 'Not enough memory allocated.'
         write(*,*) 'nBondMax: ',nBondsMax
         write(*,*) 'nBonds:   ',nBonds
      end if
c
      do i=1,nBonds
         read(*,*) (list(j),j=1,12)
         r(1)=dfloat(list(1))/dfloat(list(2))
         r(2)=dfloat(list(3))/dfloat(list(4))
         r(3)=dfloat(list(5))/dfloat(list(6))
         call Matrix_Multiply_vM(r,LatCart,HBond(1,1,i)) ! 1st Vertex
         r(1)=dfloat(list(7))/dfloat(list(8))
         r(2)=dfloat(list(9))/dfloat(list(10))
         r(3)=dfloat(list(11))/dfloat(list(12))
         call Matrix_Multiply_vM(r,LatCart,HBond(1,2,i))
      end do
c
      return
      end
c
c
      subroutine GenerateTRANS()
      implicit none
c
      include 'HBondConfigXYZ.inc'
c
      integer i,j,k,iT
      real*8 r(3)
c
      iT=0
      do i=-1,1
         r(1)=dfloat(i)
         do j=-1,1
            r(2)=dfloat(j)
            do k=-1,1
               r(3)=dfloat(k)
               iT=iT+1
               call Matrix_Multiply_vM(r,LatPBC,TRANS(1,iT))
            end do
         end do
      end do
c
      return
      end
c
c
      subroutine Matrix_Multiply_vM(v,M,r)
      implicit none
c
      real*8 M(3,3),v(3),r(3)
c
      r(1)=v(1)*M(1,1)+v(2)*M(2,1)+v(3)*M(3,1)
      r(2)=v(1)*M(1,2)+v(2)*M(2,2)+v(3)*M(3,2)
      r(3)=v(1)*M(1,3)+v(2)*M(2,3)+v(3)*M(3,3)
c
      return
      end
c
c
      subroutine GenerateHYD()
      implicit none
c
      include 'HBondConfigXYZ.inc'
c
      integer i
      real*8 ONETHIRD,TWOTHIRD
c
      ONETHIRD=1.d0/3.d0
      TWOTHIRD=2.d0/3.d0
c
      do i=1,nBonds
         if(b(i).eq.1) then
            Hyd(1,i)=TWOTHIRD*HBond(1,1,i)+ONETHIRD*HBond(1,2,i)
            Hyd(2,i)=TWOTHIRD*HBond(2,1,i)+ONETHIRD*HBond(2,2,i)
            Hyd(3,i)=TWOTHIRD*HBond(3,1,i)+ONETHIRD*HBond(3,2,i)
         elseif(b(i).eq.-1) then
            Hyd(1,i)=ONETHIRD*HBond(1,1,i)+TWOTHIRD*HBond(1,2,i)
            Hyd(2,i)=ONETHIRD*HBond(2,1,i)+TWOTHIRD*HBond(2,2,i)
            Hyd(3,i)=ONETHIRD*HBond(3,1,i)+TWOTHIRD*HBond(3,2,i)
         else
            write(*,*) 'Error: GenerateHYD'
            write(*,*) 'Unexpected Bond Variable'
            write(*,*) 'Bond: ',i,'  Variable: ',b(i)
            write(*,*) 'Defects not currently supported'
            stop
         end if
      end do
c
      return
      end
c
c
      subroutine PBCHYD()
      implicit none
c
      include 'HBondConfigXYZ.inc'
c
      integer i,j,k,nj,MinIndx(2,2)
      real*8 dist,MinDist(2)
      logical Assigned(nBondsMax)
c
      do i=1,nBonds
         Assigned(i)=.false.
      end do
c
      do i=1,nVertices
         nj=0
         do j=1,nBonds
            if(.not.Assigned(j)) then ! Skip hydrogens already placed
               nj=nj+1 
               call MinDistance(i,j,dist,k)
               if(nj.eq.1) then
                  MinIndx(1,1)=k ! Translation Index
                  MinIndx(2,1)=j ! Hydrogen index
                  MinDist(1)=dist ! Minimum distance
               elseif(nj.eq.2) then
                  if(dist.gt.MinDist(1)) then
                     MinIndx(1,2)=k ! Translation Index
                     MinIndx(2,2)=j ! Hydrogen index
                     MinDist(2)=dist ! Minimum distance
                  else
                     MinIndx(1,2)=MinIndx(1,1)
                     MinIndx(2,2)=MinIndx(2,1)
                     MinDist(2)=MinDist(1)
c
                     MinIndx(1,1)=k ! Translation Index
                     MinIndx(2,1)=j ! Hydrogen index
                     MinDist(1)=dist ! Minimum distance
                  end if
               else
                  if(dist.lt.MinDist(1)) then
                     MinIndx(1,2)=MinIndx(1,1)
                     MinIndx(2,2)=MinIndx(2,1)
                     MinDist(2)=MinDist(1)
c
                     MinIndx(1,1)=k
                     MinIndx(2,1)=j
                     MinDist(1)=dist
                  elseif(dist.lt.MinDist(2)) then
                     MinIndx(1,2)=k
                     MinIndx(2,2)=j
                     MinDist(2)=dist
                  end if
               end if
            end if
         end do
c
c -- The two closest hydrogens will have their positions updated.  They
c -- will no longer be looped over.
         k=MinIndx(2,1)         ! Hydrogen #1 Index
         Hyd(1,k)=Hyd(1,k)+TRANS(1,MinIndx(1,1))
         Hyd(2,k)=Hyd(2,k)+TRANS(2,MinIndx(1,1))
         Hyd(3,k)=Hyd(3,k)+TRANS(3,MinIndx(1,1))
         Assigned(k)=.true.
c
         k=MinIndx(2,2)         ! Hydrogen #1 Index
         Hyd(1,k)=Hyd(1,k)+TRANS(1,MinIndx(1,2))
         Hyd(2,k)=Hyd(2,k)+TRANS(2,MinIndx(1,2))
         Hyd(3,k)=Hyd(3,k)+TRANS(3,MinIndx(1,2))
         Assigned(k)=.true.
      end do
c
      return
      end
c
c
      subroutine WriteXYZ()
      implicit none
c
      include 'HBondConfigXYZ.inc'
c
      integer i,j
c
      write(2,*) nVertices+nBonds
      write(2,*) 'Config: ',nConfig
      do i=1,nVertices
         write(2,10) '  O  ',(Vertice(j,i),j=1,3)
      end do
      do i=1,nBonds
         write(2,10) '  H  ',(Hyd(j,i),j=1,3)
      end do
c
 10   format(a5,3(3x,f12.6))
c
      return
      end
c
c
      subroutine MinDistance(I,J,MinDist,MinIndx)
      implicit none
c
      include 'HBondConfigXYZ.inc'
c
      integer I,J,iT,MinIndx
      real*8 r(3),MinDist,dist,FUNC_NORM
c
      do iT=1,27
         r(1)=Vertice(1,I)-(Hyd(1,J)+TRANS(1,iT))
         r(2)=Vertice(2,I)-(Hyd(2,J)+TRANS(2,iT))
         r(3)=Vertice(3,I)-(Hyd(3,J)+TRANS(3,iT))
         dist=FUNC_NORM(r)
         if(iT.eq.1) then
            MinDist=dist
            MinIndx=iT
         else
            if(dist.lt.MinDist) then
               MinDist=dist
               MinIndx=iT
            end if            
         end if
      end do
c
      return
      end
c
c
      real*8 function FUNC_NORM(r)
      implicit none
c
      real*8 r(3)
c
      FUNC_NORM=dsqrt(r(1)*r(1)+r(2)*r(2)+r(3)*r(3))
c
      return
      end
